############################################
# Purpose: This script has multiple purposes:
#             Firstly, it can be used to check whether a code has coherency in the types
#             of subprograms call.
#
#             Secondly, it can fix the code whenever it can be solved with a cast.
#
#             Thirdly, in case we want not only to apply these fixes into the preprocessed source-code but also into
#             the original source-code, using the option --additional-source we can provide a folder with the
#             original source-code that will be modified.
#
#
#
# Inputs:
#   -v FILE, --vault = FILE
#   -s FILE, --input - source = FILE
#   --additional - source = FILE
#   -w WORKING_PRECISION, --working - precision = WORKING_PRECISION
#   -t TARGET_PRECISION, --target - precision = TARGET_PRECISION
#   -f, --fix - dependencies
############################################
from AutoRPE.UtilsRPE.SourceManager import expand_sources
from AutoRPE.UtilsRPE.SourceManager import load_vault
from AutoRPE.UtilsRPE.ImplementRPE import ImplementRPE
import AutoRPE.UtilsRPE.CurrentBlock as CurrentBlock
from AutoRPE.UtilsRPE.Inserter import find_call_in_original_file
from AutoRPE.UtilsRPE.Getter import get_real_in_line
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
from os.path import basename
import warnings


def code_check_file(filename, vault, fix=False, original_sources=None):
    """
    Check that all functions and subroutines in a source code have coherent types.
    """
    try:
        current_block = CurrentBlock.CurrentBlock(vault[filename.split("/")[-1].replace(".f90", "")].main())
    except AttributeError:
        warnings.warn("No module %s" % filename)
        return
    f = ImplementRPE(filename, None)
    if original_sources:
        original_filename = original_sources[filename]

    f.pre_process()
    # Loop through lines
    for line in f.lines:
        # Returns all REAL var on the line which corresponding dummy argument has different type
        procedure_tf, arguments_proc_tf, argument_index_tf, dummy_var_type_tf, dummy_var_intent_tf = \
            get_real_in_line(line, vault, current_block=current_block, check=True)
        # If something was found check if can be fixed
        if [procedure_tf, arguments_proc_tf, argument_index_tf, dummy_var_type_tf] != [None, None, None, None]:
            for p_tf, ap_tf, ai_tf, dv_tf, dv_i_tf in zip(procedure_tf, arguments_proc_tf, argument_index_tf,
                                                          dummy_var_type_tf, dummy_var_intent_tf):
                types = {"single": "sp", "double": "dp"}
                try:
                    target_type = types[dv_tf]
                except KeyError:
                    continue
                # If the intent is inout or out it can not be fixed with a cast
                if fix:
                    if dv_i_tf in ["out", "inout"]:
                        fix_this = False
                    else:
                        fix_this = True
                else:
                    fix_this = False

                fixed = find_call_in_original_file(p_tf, ap_tf, ai_tf, filename,
                                                   target_type=target_type, fix=fix_this)
                if original_sources:
                    original_fixed = find_call_in_original_file(p_tf, ap_tf, ai_tf, original_filename,
                                                                target_type=target_type, fix=fix_this)


                if fixed:
                    if fix:
                        print("Was fixed!")
                    else:
                        print("Can be fixed!")
                else:
                    print("Can't be fixed")

        else:
            # If not, look if the code is entering or exiting a procedure block
            current_block = current_block.in_which_block(line, vault=vault)


if __name__ == "__main__":
    # Getting Command line options
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option("-v", "--vault", dest="vault", default=None, metavar="FILE")
    parser.add_option("-s", "--input-source", dest="input_source", default=None, metavar="FILE")
    parser.add_option("-w", "--working-precision", dest="working_precision", default="single")
    parser.add_option("-t", "--target-precision", dest="target_precision", default="double")
    parser.add_option("-f", "--fix-dependencies", dest="fix_dependencies", default=False, action="store_true")
    parser.add_option("--additional-source", dest="additional_source", default=None, metavar="FILE")

    (options, args) = parser.parse_args()
    vault_path = options.vault
    VariablePrecision.wp = options.working_precision
    target_precision = options.target_precision
    path_to_input_sources = options.input_source
    additional_sources_path = options.additional_source
    fix_dependencies = options.fix_dependencies

    if path_to_input_sources is None or vault_path is None:
        print("\nArguments -s/--input-source and -v/--vault are mandatory\n")
        parser.print_help()
        exit(1)

    # Loading vault
    vault = load_vault(vault_path)

    # Find all files in sources
    all_sources = expand_sources(path_to_input_sources)

    # In case the argument additional sources is provided,
    # create a dictionary to know which file from this folder corresponds to each preprocessed file.
    if additional_sources_path is not None:
        additional_sources = expand_sources(additional_sources_path)

        def match_sources(preprocessed, original):
            dictionary = {}
            for file in preprocessed:
                F90_filename = basename(file).replace(".f90", ".F90")
                for f in original:
                    if basename(f) == F90_filename:
                        dictionary[file] = f
                        break
            return dictionary
        dictionary_of_additional_sources = match_sources(all_sources, additional_sources)
    else:
        dictionary_of_additional_sources = None

    if not all_sources:
        parser.print_help()
        print("\nSources folder provided does not contain source files.\n")
        exit(1)
    # Process Sources!
    total = len(all_sources)
    for index, filename in enumerate(all_sources):
        code_check_file(filename, vault, fix=fix_dependencies, original_sources=dictionary_of_additional_sources)
